use base64::{engine::general_purpose, Engine as _};
use serde::{
    ser::{self, Serializer},
    Deserialize, Serialize,
};
use std::collections::HashMap;
use thiserror::Error;

pub mod profile;

#[derive(Error, Debug)]
pub enum Error {
    #[error("Issue at the serializer: {0}")]
    Json(#[from] serde_json::Error),

    #[error("Base64: {0}")]
    Base64(#[from] base64::DecodeError),
}

#[derive(Eq, PartialEq, Clone, Debug, Serialize, Deserialize)]
pub struct RelaySettings {
    pub read: bool,
    pub write: bool,
}

#[derive(Eq, PartialEq, Clone, Debug)]
pub struct EncryptedData {
    pub encrypted_message: Vec<u8>,
    pub iv: Vec<u8>,
}

#[derive(Eq, PartialEq, Clone, Debug)]
pub enum Content {
    Profile(profile::Profile),
    Text(String),
    RecommendServer(String),
    ContactList(HashMap<String, RelaySettings>),
    DirectMessage(EncryptedData),
    Unknown(u32, String),
}

impl ser::Serialize for Content {
    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
    where
        S: Serializer,
    {
        serializer.serialize_str(
            &self
                .try_to_string()
                .map_err(|e| ser::Error::custom(e.to_string()))?,
        )
    }
}

impl Content {
    pub fn kind(&self) -> u32 {
        match self {
            Self::Profile(_) => 0,
            Self::Text(_) => 1,
            Self::RecommendServer(_) => 2,
            Self::ContactList(_) => 3,
            Self::DirectMessage(_) => 4,
            Self::Unknown(k, _) => *k,
        }
    }

    pub fn deserialize(kind: u32, content: &str) -> Result<Self, Error> {
        match kind {
            0 => Ok(Self::Profile(serde_json::from_str(content)?)),
            1 => Ok(Self::Text(content.to_owned())),
            2 => Ok(Self::RecommendServer(content.to_owned())),
            3 => Ok(Self::ContactList(serde_json::from_str(content)?)),
            4 => {
                let parts = content.split("?iv=").collect::<Vec<&str>>();
                if parts.len() == 2 {
                    let encrypted_message = general_purpose::STANDARD.decode(&parts[0])?;
                    let iv = general_purpose::STANDARD.decode(&parts[0])?;

                    Ok(Self::DirectMessage(EncryptedData {
                        encrypted_message,
                        iv,
                    }))
                } else {
                    Ok(Self::Unknown(4, content.to_owned()))
                }
            }
            _ => Ok(Self::Unknown(kind, content.to_owned())),
        }
    }

    pub fn try_to_string(&self) -> Result<String, Error> {
        match self {
            Self::Profile(p) => Ok(serde_json::to_string(&p)?),
            Self::ContactList(p) => Ok(serde_json::to_string(&p)?),
            Self::DirectMessage(message) => {
                let encrypted_message =
                    general_purpose::STANDARD.encode(&message.encrypted_message);
                let iv = general_purpose::STANDARD.encode(&message.iv);
                let to_encode = format!("{}?iv={}", encrypted_message, iv);
                Ok(serde_json::to_string(&to_encode)?)
            }
            Self::Text(t) | Self::RecommendServer(t) | Self::Unknown(_, t) => Ok(t.to_owned()),
        }
    }
}
